//! Air.cpp

#include <omnetpp.h>
#include "Node.h"
#include "MAC.h"
#include "MessageTypes.h"
#include "debug.h"
#include "Observer.h"

#define _CHECK__NEIGHBOURS

//! relation between screen units and meters
#define SCRTOREAL( scr_dist ) (scr_dist / 10)

//! minimum distance allowed between nodes; in meters
#define MIN_DIST 0,1

//
//
// Radio propagation model

#define no_Xsigma_Pr( d ) (Pt + Gt - ( PLd0 + 10 * n * log10((double)d/d0) ) + Gr)
#define Pr( d ) (Pt + Gt - ( PLd0 + 10 * n * log10((double)d/d0) + normal(0, sigma) ) + Gr)
#define RSSIval( d ) (RSSIoff + Pr(d))				

//!the transmit power
const int Pt = 0;
//!the transmit antenna gain
const int Gt = 0;
//!the receive antenna gain
const int Gr = 0;
//!reference distance d0
const int d0 = 1;
//!the free-space path loss at a reference distance d0
const double PLd0 = 40.04599702;
//!the path loss exponent (which depends on the propagation environment)
const double n = 2.5;
//!standard deviation of the shadowing term Xs
const int sigma = 5;
//!RSSI offset 
const int RSSIoff = 40; //35
//const int RSSIoff = 60;
//! RSSI receive threshold for CCA
const int CCA_THR = (-32);

//! Helper class, to store neighbour nodes and respective distance
/*
	\author	Nuno Pereira
	\date Act 2005.
*/
class Neighbour
{
public:
 	  Neighbour(int nID, double d) {
		  nodeID = nID;
		  dist = d;
	  }

public:
	   int 		nodeID;
	   double	dist;
};

//! Air simple module class definition
/*

	\author	Nuno Pereira
	\date Act 2005.
*/
class Air : public cSimpleModule
{
// Member variables
private:
	// retrieved module parameters and pre-calculated data

	//! retrieved module parameter; alfa: time-of-flight upper bound
	cPar*		alfa;

	//! retrieved module parameter; probability that a carrier is *not* sent 
	double 		pError;
	
	//! reference to module used to observe results of the tourmaments 
	Observer *observer;
	
	long carrierErrorCounter;
	long carrierCounter;
	
public:
	// public variables
	
	//! list of neighbouring nodes (within maximum radio range)
	std::vector<Neighbour> neighbours;

	//! list of two hops neighbouring nodes (within maximum radio range)
	std::vector<Neighbour> twoHopNeighbours;
	
public:
    Module_Class_Members(Air,cSimpleModule,0)
    
    //! Do initialization of class data members in more than one (two) stage
    virtual int numInitStages () const;
	//! Do initialization of class data members and any other initialization procedure necessary
    virtual void initialize(int stage);
	//! Process messages received.
    virtual void handleMessage(cMessage *msg);
	//! Saves summary results (if any)
    virtual void finish();

	virtual ~Air();

// Aditional methods
private:
	bool insertTwoHopNeighbour(int id, int dist);
	void sendNeighbours(cMessage *msg);
	
	bool checkNeighbours(std::vector<Neighbour> &neighbours, int ourPrio);
	bool checkNeighbourNeighbours(std::vector<Neighbour> &neighbours, int NeighbPrio, int ourID);
};

Define_Module( Air );

//! Do initialization of class data members in two stages
int Air::numInitStages() const
{
	return 2; 
}

// returns true if two hop neighbour is not a duplicate, false otherwise
bool Air::insertTwoHopNeighbour(int id, int dist) {
	std::vector<Neighbour>::iterator iter;

	// two hop neighbour is not in neighbour list
  for ( iter = neighbours.begin( ) ; iter != neighbours.end( ) ; iter++ )
		if ((*iter).nodeID == id) break;	
	if ((*iter).nodeID == id) return FALSE; // do not insert duplicates

	// two hop neighbour is not already in two hop neighbour list
	if (twoHopNeighbours.size() == 0) { 
		twoHopNeighbours.insert(twoHopNeighbours.begin(), Neighbour(id, dist));
		return TRUE;
	}
   	for ( iter = twoHopNeighbours.begin( ) ; iter != twoHopNeighbours.end( ) ; iter++ )
		if ((*iter).nodeID == id) break;
	if ((*iter).nodeID == id) return FALSE; // do not insert duplicates

	twoHopNeighbours.insert(twoHopNeighbours.begin(), Neighbour(id, dist));
	return TRUE;
}

//!Do initialization of class data members and any other initialization procedure necessary
void Air::initialize(int stage)
{
	//! get reference to containing node
	cModule *ourNode = parentModule();
	ASSERT(dynamic_cast<Node *>(ourNode) != NULL);

	if (stage == 0) { // first init stage
		// retrieve module parameters
		alfa = &par("alfa");
		pError = ((double) par("pError"))/* * 10000000000*/;
		ev << "pError=" << pError;
		carrierCounter=0;
		carrierErrorCounter=0;
		
		WATCH(carrierCounter);
		WATCH(carrierErrorCounter);
		
		//
		// build neighbours list
	
		// determine maximum range of the radio; in meters (used to build neighbour list)
		// this depends on radio propagation model parameters
		int RADIO_RANGE=1;
		while ((RSSIoff + no_Xsigma_Pr(RADIO_RANGE++)) > CCA_THR); // + sigma, to account for max of variability; + 1, to be safe 
		//std::cout << "RADIO_RANGE:" << RADIO_RANGE << "\n";
		 		
		// position of the containing node
		double our_px = (double) ourNode->par("px");
		double our_py = (double) ourNode->par("py");
	
		// get system module
		cModule *sysModule = ourNode;
		while (sysModule->parentModule() != NULL) sysModule = sysModule->parentModule();
		ASSERT(sysModule != NULL);
		ev << "\nneighbours(" << ourNode->index() << ") => {\n";
		// find all nodes in the network and see if their position is close enough
		for (cSubModIterator modIter(*sysModule); !modIter.end(); modIter++)
		{
			cModule *module = modIter();
			if (dynamic_cast<Node *>(module) != NULL) {
				Node *node = (Node *) module;
				if (node != ourNode) {						
					// position of the node
					double px = (double) node->par("px");
					double py = (double) node->par("py");
		
					double dx = our_px - px;
					double dy = our_py - py;
					double dist = SCRTOREAL(sqrt(dx*dx + dy*dy));
									
					ASSERT(dist>=MIN_DIST);
					
					if (dist < RADIO_RANGE) {
						ev << "\t\t" << node->index() << "\t(d=" << dist << ")\n";
						neighbours.push_back(Neighbour(node->id(), dist));			
					}
				}
			}
		}
		ev << "} ";
		ev << ourNode->index() << " :: "; 
		if (neighbours.size() == ((int)sysModule->par("numNodes"))-1) ev << "CONECTED WITH ALL OTHER NODES;\n"; 
		else {
			ev << "*NOT* CONECTED WITH ALL OTHER NODES;\n";
	//		std::cout << ourNode->index() << " :: " << "*NOT* CONECTED WITH ALL OTHER NODES;\n";
		}
		
		observer = (Observer*) simulation.moduleByPath ("observer");
		ASSERT(observer != NULL);
 
	} else if (stage == 1) { // second init stage
		//
		// build two hop neighbours list
		if (neighbours.size() == 0) return;
		
		ev << "\nTwo hop neighbours(" << ourNode->index() << ") => {\n";
		for (int i=0; i<neighbours.size(); i++) {		
			Air *neighbourAir = ((Air*)(simulation.module(neighbours[i].nodeID)->submodule("air")));
			ASSERT(neighbourAir!=NULL);
			std::vector<Neighbour> neighbourNeighbours = neighbourAir->neighbours;
			for (int j=0; j<neighbourNeighbours.size(); j++) {
				if (ourNode->id() != neighbourNeighbours[j].nodeID) {
					if (insertTwoHopNeighbour(neighbourNeighbours[j].nodeID, neighbours[j].dist+neighbourNeighbours[j].dist)) {
						neighbourAir = ((Air*)(simulation.module(neighbourNeighbours[j].nodeID)->submodule("air")));
						ev << "\t\t" << neighbourAir->parentModule()->index() << "\t(d=" << neighbours[j].dist+neighbourNeighbours[j].dist << ")\n";
					}
				}				
			}
		}
		ev << "} \n";
	}
}

//!	Process messages received.
/*!
	\param msg - cMessage* : Message received.
*/
void Air::handleMessage(cMessage *msg)
{
	//std::cout << parentModule()->index() << "[" << simTime() << "] - Air:: messageKind:" << msg->kind() << "\n";
	if (msg->arrivedOn("from_radio") == true) {	
		sendNeighbours(msg);
#ifdef _CHECK__NEIGHBOURS
		// if message is a data message tx
		// check every two hop neighbours if they where listenning
		// or had something to send.
		// in the latter case, we check the neighbour priority
		if (msg->kind() == MSG_TYPE_DATA_MSG) {
			// check one hop neighbours
			if (checkNeighbours(neighbours, msg->priority()) == FALSE) {
				/////////////////////////////////////
				// say to observer that a prioritization error occurred 
				observer->prioError();
				/////////////////////////////////////
			}

			// check two hops neighbours
			if (checkNeighbours(twoHopNeighbours, msg->priority()) == FALSE) {
				/////////////////////////////////////
				// say to observer that a prioritization error occurred 
				observer->prioError();
				/////////////////////////////////////
			}
			
		}
#endif // _CHECK_NEIGHBOURS	
		delete msg;
	} else { // message received from remote node, calculate rssi
	/* 
		double rssi = RSSIval(msg->par("dist"));
		msg->addPar("rssi");
		msg->par("rssi") = rssi;
	*/		
		send(msg, "to_radio");	
	} 
}

bool Air::checkNeighbours(std::vector<Neighbour> &neighbours, int ourPrio) {
	// check one hop neighbours
	for (int i=0; i<neighbours.size(); i++) {
		MAC *neighbourMAC = ((MAC*)(simulation.module(neighbours[i].nodeID)->submodule("mac")));
		ASSERT(neighbourMAC!=NULL);
			 
		if (neighbourMAC->lastPrio() != INVALID_PRIO) { // check the neighbour priority, if he participated in the tournament
			// our message priority is in the message priority parameter
			if (neighbourMAC->lastPrio() < ourPrio) {			
				Air *neighbourAir = ((Air*)(simulation.module(neighbours[i].nodeID)->submodule("air")));
				ASSERT(neighbourAir!=NULL);
				// check neigbour neigbours
				bool checkNN=checkNeighbourNeighbours(neighbourAir->neighbours, neighbourMAC->lastPrio(), parentModule()->id());
				// check neigbour: two hop neigbours
				checkNN = checkNN || checkNeighbourNeighbours(neighbourAir->twoHopNeighbours, neighbourMAC->lastPrio(), neighbours[i].nodeID);
				if (checkNN == FALSE) {
					printf("\n%d: Prio verification failed for neighbour %d (Prio=%d < %d)\n", parentModule()->index(), neighbourMAC->parentModule()->index(), neighbourMAC->lastPrio(), ourPrio);
//					ASSERT(FALSE);					
					return FALSE;
				}
			}
		}				
	}
	return TRUE;
}			

bool Air::checkNeighbourNeighbours(std::vector<Neighbour> &neighbours, int NeighbPrio, int ourID) {
	for (int i=0; i<neighbours.size(); i++) {
		if (neighbours[i].nodeID != ourID) {
			MAC *neighbourMAC = ((MAC*)(simulation.module(neighbours[i].nodeID)->submodule("mac")));
			ASSERT(neighbourMAC!=NULL);
			if (neighbourMAC->lastPrio() != INVALID_PRIO) return TRUE;
			if (neighbourMAC->lastPrio() != INVALID_PRIO) { // check the neighbour priority, if he participated in the tournament				
				// our message priority is in the message priority parameter
				if (neighbourMAC->lastPrio() < NeighbPrio) {
					return TRUE;
				}
			} else return TRUE;
		}				
	}
	return FALSE;	
}

void Air::sendNeighbours(cMessage *msg) {
	bool bError=FALSE;

	if (neighbours.size() == 0) return;
//	printf("SENDING TO:");
	for (int i=0; i<neighbours.size(); i++) {
		// we put the distance on the message, to make things easier at the receiver...
//		msg->addPar("dist");
//		msg->par("dist") = neighbours[i].dist;

		// only introduce tx errors at the start of sending carriers		
		if (msg->kind() == MSG_TYPE_RADIO_CARRIER_ON && pError > 0) {
			carrierCounter++;
			//double p = rng(0)->doubleRandIncl1(); // Random double on the [0,1] interval.
			//long p = intrand(10000000001);
			if (bernoulli(pError)) {
				//std::cout << "\n*\npError=" << pError << "; \n*\n";
				carrierErrorCounter++;
				bError = TRUE;
			}
		} 		
		
		if (bError == FALSE) {
			// send message to destination
			sendDirect((cMessage *)msg->dup(), (double)*alfa, simulation.module(neighbours[i].nodeID)->gate("in"));
		}
	}
//	printf("\n");
}

//! Saves summary results (if any)
void Air::finish()
{	
    // save summary results (if any)
/*    
    ev << endl;
    ev << "total carriers: " << carrierCounter << endl;
    ev << "total carrier errors: " << carrierErrorCounter << endl;

    recordScalar("total carriers: ", carrierCounter);
    recordScalar("total carrier errors: ", carrierErrorCounter);
*/	    
}

Air::~Air()
{	
}
